﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using LibUsbDotNet.Main;
using LibUsbDotNet;
using MonoLibUsb;
using System.Threading;

namespace calcout {
	static class Calcout {
		class Calc {
			public class CalcNotFoundException : Exception { public CalcNotFoundException(string message) : base(message) { } }
			UsbDeviceFinder Finder84P = new UsbDeviceFinder(0x0451, 0xE003);
			UsbDeviceFinder Finder84PSE = new UsbDeviceFinder(0x0451, 0xE008);

			UsbDevice calc;
			Queue<byte[]> dataBack = new Queue<byte[]>();
			public ErrorCode ErrorCode { get; set; }

			private bool hasData { get { return dataBack.Count != 0; } }

			UsbEndpointReader reader;
			UsbEndpointWriter writer;

			public bool IsOpen { get { return calc.IsOpen; } }

			public Calc(System.EventHandler<EndpointDataEventArgs> OnRecieve) {
				calc = UsbDevice.OpenUsbDevice(Finder84P);
				if (calc == null) {
					calc = UsbDevice.OpenUsbDevice(Finder84PSE);
					if (calc == null) {
						throw new CalcNotFoundException("Calculator not found.");
					}
				}

				IUsbDevice wholeUsbDevice = calc as IUsbDevice;
				if (!ReferenceEquals(wholeUsbDevice, null)) {
					// This is a "whole" USB device. Before it can be used, 
					// the desired configuration and interface must be selected.

					// Select config #1
					wholeUsbDevice.SetConfiguration(1);

					// Claim interface #0.
					wholeUsbDevice.ClaimInterface(0);
				}

				reader = calc.OpenEndpointReader(ReadEndpointID.Ep01);
				writer = calc.OpenEndpointWriter(WriteEndpointID.Ep02);
				reader.DataReceived += (OnRecieve);
				reader.DataReceivedEnabled = true;
				Console.WriteLine(calc.Info.Descriptor.ToString());
			}

			public int SendData(ICollection<byte> data) {
				int bytesWritten;
				ErrorCode = writer.Write(data.ToArray(), 2000, out bytesWritten);
				Console.WriteLine(ErrorCode);
				return bytesWritten;
			}

			public void Disconnect() {
				if (calc != null) {
					if (calc.IsOpen) {
						// If this is a "whole" usb device (libusb-win32, linux libusb-1.0)
						// it exposes an IUsbDevice interface. If not (WinUSB) the 
						// 'wholeUsbDevice' variable will be null indicating this is 
						// an interface of a device; it does not require or support 
						// configuration and interface selection.
						IUsbDevice wholeUsbDevice = calc as IUsbDevice;
						if (!ReferenceEquals(wholeUsbDevice, null)) {
							// Release interface #0.
							wholeUsbDevice.ReleaseInterface(0);
						}
						calc.Close();
					}
					calc = null;

					// Free usb resources
					UsbDevice.Exit();
				}
			}
		}

		static Queue<byte[]> DataQueue;

		static void Main(string[] args) {
			Calc c = null;
			try {
				c = new Calc(OnUSBData);
			} catch (Calc.CalcNotFoundException) {
				PrintErrorAndExit("Could not connect to calculator, please make sure it is turned on and connected.", -3);
			} catch (Exception e) {
				PrintErrorAndExit("Could not connect to calculator: " + e.ToString(), -2);
			}

			DataQueue = new Queue<byte[]>();
			StringBuilder sb = new StringBuilder();
			while (true) {
				if (Console.KeyAvailable) {
					var key = Console.ReadKey(true);
					if (key.Key == ConsoleKey.Escape) {
						c.Disconnect();
						Environment.Exit(0);
					} else if (key.Key == ConsoleKey.Enter) {
						Console.WriteLine();
						try {
							c.SendData(GetByteArray(sb.ToString()));
						} catch {
							Console.WriteLine("Error in data.");
						}
						sb.Clear();
					} else if (key.Key == ConsoleKey.Backspace) {
						if (sb.Length > 0) {
							sb.Remove(sb.Length - 1, 1);
							Console.Write(key.KeyChar);
							Console.Write(' ');
							Console.Write(key.KeyChar);
						}
					} else {
						sb.Append(key.KeyChar);
						Console.Write(key.KeyChar);
					}
				}
				while (DataQueue.Count != 0) {
					Console.WriteLine(HexDump(DataQueue.Dequeue()));
				}
				Thread.Sleep(100);
			}
		}

		private static void OnUSBData(object sender, EndpointDataEventArgs e) {
			DataQueue.Enqueue(e.Buffer.SubArray(0, e.Count));
		}

		/// <summary>
		/// Converts the passed in string of hex characters to an array of bytes.
		/// </summary>
		/// <param name="s">The string to convert. Must be a hex string, can be prefaced with "0x" or "$", or end with "h".</param>
		/// <param name="arrayLen">The desired length of the returned array, will zero fill the end of the array.</param>
		/// <returns>An array of bytes corresponding to the hex string passed in.</returns>
		/// <example>GetByteArray("0xFFFF") => {0xFF, 0xFF}</example>
		public static byte[] GetByteArray(string s) {
			if (s.StartsWith("0x"))
				s = s.Remove(0, 2);
			if (s.StartsWith("$"))
				s = s.Remove(0, 1);
			if (s.EndsWith("h"))
				s = s.Remove(s.Length - 1);
			int n = s.Length;
			List<byte> bytes = new List<byte>();
			for (int i = 0; i < n; i++) {
				if (s[i] == '[') {
					int end = s.IndexOf(']', i);
					bytes.AddRange(ASCIIEncoding.ASCII.GetBytes(s.Substring(i + 1, end - 1 - i)));
					i = end;
				} else {
					try {
						bytes.Add(Convert.ToByte(s.Substring(i, 2), 16));
					} catch {
						//failed to convert these 2 chars, they may contain illegal characters
						bytes.Add(0);
					}
					i++;
				}
			}
			return bytes.ToArray();
		}

		public static string HexDump(byte[] bytes) {
			if (bytes == null) return "<null>";
			int len = bytes.Length;
			StringBuilder result = new StringBuilder(((len + 15) / 16) * 78);
			char[] chars = new char[78];
			// fill all with blanks
			for (int i = 0; i < 75; i++) chars[i] = ' ';
			chars[76] = '\r';
			chars[77] = '\n';

			for (int i1 = 0; i1 < len; i1 += 16) {
				chars[0] = HexChar(i1 >> 28);
				chars[1] = HexChar(i1 >> 24);
				chars[2] = HexChar(i1 >> 20);
				chars[3] = HexChar(i1 >> 16);
				chars[4] = HexChar(i1 >> 12);
				chars[5] = HexChar(i1 >> 8);
				chars[6] = HexChar(i1 >> 4);
				chars[7] = HexChar(i1 >> 0);

				int offset1 = 11;
				int offset2 = 60;

				for (int i2 = 0; i2 < 16; i2++) {
					if (i1 + i2 >= len) {
						chars[offset1] = ' ';
						chars[offset1 + 1] = ' ';
						chars[offset2] = ' ';
					} else {
						byte b = bytes[i1 + i2];
						chars[offset1] = HexChar(b >> 4);
						chars[offset1 + 1] = HexChar(b);
						chars[offset2] = (b < 32 ? '·' : (char)b);
					}
					//offset1 += (i2 == 8 ? 4 : 3);
					offset1 += (i2 == 7 ? 4 : 3);
					offset2++;
				}
				result.Append(chars);
			}
			return result.ToString();
		}

		private static char HexChar(int value) {
			value &= 0xF;
			if (value >= 0 && value <= 9) return (char)('0' + value);
			else return (char)('A' + (value - 10));
		}

		public static T[] SubArray<T>(this T[] data, int index, int length) {
			T[] result = new T[length];
			Array.Copy(data, index, result, 0, length);
			return result;
		}

		private static void PrintErrorAndExit(string message, int errorCode) {
			Console.Error.WriteLine(message);
#if DEBUG
			Console.ReadLine();
#endif
			Environment.Exit(errorCode);
		}
	}
}
